/* \file functions.c
 * \brief \A Advanced functions to control a Maxon motor with an EPOS control unit. <br>
 *
*/

/* \file functions.c
 * \brief Advanced functions to control a Maxon motor with an EPOS control unit.
 *
*/

/*! \file functions.c
 * \brief Advanced functions to control a Maxon motor with an EPOS control unit.<br>
 *
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/file.h>
#include <sys/select.h>

#include <sys/msg.h>
#include <time.h>

#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <pthread.h>
#include <sys/msg.h>

#include "msg_queue.h"
#include "epos.h"
#include "functions.h"
#include "init_parameter.h"
#include "global_def.h"


#define HOMING_MODE                             6    ///< mode of operation: homing mode
#define PROFILE_POSITION_MODE                   1    ///< mode of operation: profiled position mode
#define POSITION_MODE                          -1    ///< mode of operation: direct position mode

#define ACCELERATION_POSITION_MODE             12    ///< acceleration for profiled position mode for the centrifuges
#define DECELERATION_POSITION_MODE             12    ///< deceleration for profiled position mode for the centrifuges
#define VELOCITY_POSITION_MODE                 12    ///< velocity for profiled position mode for the centrifuges

#define MAX_VELOCITY_DIRECT_POS_MODE       0x005f    ///< maximal velocity for direct position mode
#define MAX_ACCELAERATION_DIRECT_POS_MODE  0x00ff    ///< maximal acceleration for direct position mode

#define ACCELERATION_POSITION_MODE_CAM         10    ///< acceleration for profiled position mode for the camera motors
#define DECELERATION_POSITION_MODE_CAM         10    ///< deceleration for profiled position mode for the camera motors
#define VELOCITY_POSITION_MODE_CAM             10    ///< velocity for profiled position mode for the camera motors

#define HOMING_METH_POS_SPEED_AND_INDEX         7    ///< homing method, positive speed & index used for the centrifuges
#define HOMING_METH_NEG_SPEED_AND_INDEX        11    ///< homing method, negative speed & index used for the camera motor



pthread_mutex_t mutex_uart = PTHREAD_MUTEX_INITIALIZER;    ///< initialize mutex to protect UART


int check_if_stop_MQ_for_timeout(WORD timeout, struct motor_args *arguments, struct msg_queue *queue, time_t start_time, WORD centrifugation_time){
/*! \brief Checks message queue for a quick stop- or getTime command.
 * \param WORD timeout    : number of 100ms loops
 * \param struct motor_args *arguments  : specific motor parameters
 * \param struct msg_queue *queue  : message queue ID and local address
 * \param time_t start_time  : start time of the centrifugation process
 * \param WORD centrifugation_time  : expected centrifugation time
 * \retval   0 : no quickstop appeared
 * \retval -1 : quickstop occurred
 */
    struct msg_buf MQ_message_struct;

    unsigned char stop[4] = {'s','t','o','p'};
    unsigned char ask_for_time[4] = {'g','e','t','t'};

    WORD data[2]= {0x0000, 0x0000};
    WORD ControlwordCommand=0;
    int i;
    int received;
    time_t time_left;
    //this loops delay is approximately 100ms.
    for (i=0;i<timeout;i++){

        received=recv_message_timeout(queue, &MQ_message_struct, 50000);

        if(received != -1) {

            if(memcmp(stop, MQ_message_struct.cmd_id, 4) == 0 ){
#ifdef DEBUG_INFO_BASIC
    printf("do quick stop   addr.: %d \n", arguments->my_queue_addr);
#endif
                // write the quick stop command to the EPOS register
                ControlwordCommand=ControlwordCommand & 0x7B;
                ControlwordCommand=ControlwordCommand | 0x02;
                data[0]= ControlwordCommand;
                data[1]= 0x00;
                WriteObject(0x6040, 0x00, data, arguments);
                MQ_message_struct.param[0] = STOPPED;
                MQ_message_struct.cmd_param_len=1;
#ifdef DEBUG_INFO_BASIC
    printf("send quick stop done response  addr.: %d\n",arguments->my_queue_addr);
#endif
                send_message(queue, &MQ_message_struct, ADDR_COMD);
                return (-1);
            }else{
            if(memcmp(ask_for_time, MQ_message_struct.cmd_id, 4) == 0){
#ifdef DEBUG_INFO_BASIC
                printf("asked for time   addr.: %d \n", arguments->my_queue_addr);
#endif
                time_left=centrifugation_time-difftime(time(NULL), start_time);
                if(time_left<0){time_left=0; }
                //write the time that is left into the message queue
                MQ_message_struct.cmd_param_len=3;
                MQ_message_struct.param[0] = 'T';
                MQ_message_struct.param[1] = time_left & 0x00ff;
                MQ_message_struct.param[2] = (time_left & 0xff00)>>8;
                send_message(queue, &MQ_message_struct, ADDR_COMD);
                }else{
                        //if other message send BUSY
                        MQ_message_struct.param[0] = BUSY;
                        MQ_message_struct.cmd_param_len=1;
                        send_message(queue, &MQ_message_struct, ADDR_COMD);
                    }
            }
        }
        usleep(31000);
    }
    return(0);
}

WORD get_statusword(struct motor_args *arguments){
/*! \brief Reads the statusword from an EPOS control unit.
 * \param struct motor_args *arguments  : specific motor parameters
 * \retval WORD :  statusword
 */
    WORD *answer_buf;
    WORD answer;

    pthread_mutex_lock(&mutex_uart);
    ReadObject(0x6041, 0x00, &answer_buf, arguments);
    answer=answer_buf[3];
    pthread_mutex_unlock(&mutex_uart);
    return(answer);
}


int go_to_velocity(unsigned int acc, unsigned int dec, unsigned int velocity, struct motor_args *arguments, struct msg_queue *queue, time_t start_time, WORD centrifugation_time){
/*! \brief Drives the motor to the desired velocity.
 * \param unsigned int acc  : acceleration to reach velocity
 * \param unsigned int dec  : deceleration to reach velocity
 * \param unsigned int velocity  : velocity to reach
 * \param struct motor_args *arguments  : specific motor parameters
 * \param struct msg_queue *queue  : address from the actual message queue
 * \param time_t start_time  : start time of the centrifugation process
 * \param WORD centrifugation_time  : expected centrifugation time
 * \retval -1 : quick stop command appeared
 * \retval  0 : everything was OK
 */
#define PROFILE_VELOCITY_MODE    0x03 ///< \brief mode of operation profiled velocity mode

    WORD data[2]= {0x0000, 0x0000};
    WORD ControlwordCommand=0;
    WORD statusword;

    //write mode of operation
    data[0]= PROFILE_VELOCITY_MODE;
    data[1]= 0x0000;
    WriteObject(0x6060, 0x00, data, arguments);

    //write profile acceleration
    data[0]= acc;
    data[1]= 0x0000;
    WriteObject(0x6083, 0x00, data, arguments);

    //write profile deceleration
    data[0]= dec;
    data[1]= 0x0000;
    WriteObject(0x6084, 0x00, data, arguments);

    //write target profile velocity
    data[0]= velocity;
    data[1]= 0x0000;
    WriteObject(0x60ff, 0x00, data, arguments);

    //enable operation and execute the motion
    ControlwordCommand=ControlwordCommand & 0xfe7F;
    ControlwordCommand=ControlwordCommand | 0x000f;
    data[0]= ControlwordCommand;
    data[1]= 0x00;
    WriteObject(0x6040, 0x00, data, arguments);

    //wait until demand velocity is reached
    do{
        if (check_if_stop_MQ_for_timeout(10, arguments, queue, start_time, centrifugation_time)==-1   ){ // if quick_stop command appeared return -1
            return(-1);
        }
        statusword=get_statusword( arguments);
    }while( ( (statusword & 0x0400)>>10 )!= 1  );

    return(0);
}


int go_to_home_position(struct motor_args *arguments, struct msg_queue *queue){
/*! \brief Drives the motors to the home position.
 * \param struct motor_args *arguments  : specific motor parameters
 * \param struct msg_queue *queue  : address from the actual message queue
 * \retval -1 : quick stop command appeared
 * \retval  0 : everything worked fine
 */
    WORD data[2]= {0x0000, 0x0000};
    WORD ControlwordCommand=0;
    WORD statusword;

        //write Mode of Operation
    data[0]= HOMING_MODE;
    data[1]= 0x0000;
    WriteObject(0x6060, 0x00, data, arguments);

    go_to_defined_state(arguments);

    if( arguments->my_queue_addr != ADDRESS_CAMERA_MOTOR ){
            //write homing method
            data[0]=HOMING_METH_POS_SPEED_AND_INDEX;  // 23 is right at the light barrier
            data[1]= 0x0000;
            WriteObject(0x6098, 0x00, data, arguments);
    }else{
            //Homing mode camera motor
            data[0]=HOMING_METH_NEG_SPEED_AND_INDEX;
            data[1]= 0x0000;
            WriteObject(0x6098, 0x00, data, arguments);
        }

        //write control word--> enable operation    //state transition 4,16
        ControlwordCommand=ControlwordCommand & 0xff7F;
        ControlwordCommand=ControlwordCommand | 0x000f;
        data[0]= ControlwordCommand;
        data[1]= 0x00;
        WriteObject(0x6040, 0x00, data, arguments);

        //write control word--> start homing mode
        ControlwordCommand=ControlwordCommand & 0xfe7F;
        ControlwordCommand=ControlwordCommand | 0x0010;
        data[0]= ControlwordCommand;
        data[1]= 0x00;
        WriteObject(0x6040, 0x00,  data, arguments);

        //wait until target reached
        do{
            if( check_if_stop_MQ_for_timeout(100, arguments, queue, 0, 0)==-1   ){ // if quick_stop command appeared return -1
            return(-1);
            }
            statusword=get_statusword(arguments);
        }while( ( (statusword & 0x0400)>>10 )!= 1  );

        //write velocity and acceleration for position mode
        if( arguments->my_queue_addr != ADDRESS_CAMERA_MOTOR ){
#ifdef DIRECT_POSITION_MODE
            //write position mode
            data[0]= POSITION_MODE;
            data[1]= 0x0000;
            WriteObject(0x6060, 0x00, data, arguments);

            //write max velocity for position_mode
            data[0]= MAX_VELOCITY_DIRECT_POS_MODE;
            data[1]= 0x0000;
            WriteObject(0x607f, 0x00, data, arguments);

            //write max acceleration
            data[0]= MAX_ACCELAERATION_DIRECT_POS_MODE;
            data[1]= 0;
            WriteObject(0x60C5, 0x00, data, arguments);
#else
            //write profiled position mode
            data[0]= PROFILE_POSITION_MODE;
            data[1]= 0x0000;
            WriteObject(0x6060, 0x00, data, arguments);

            //write profile acceleration
            data[0]= ACCELERATION_POSITION_MODE;
            data[1]= 0x0000;
            WriteObject(0x6083, 0x00, data, arguments);

            //write profile deceleration
            data[0]= DECELERATION_POSITION_MODE;
            data[1]= 0x0000;
            WriteObject(0x6084, 0x00, data, arguments);

            //write profile velocity
            data[0]= VELOCITY_POSITION_MODE;
            data[1]= 0x0000;
            WriteObject(0x6081, 0x00, data, arguments);
#endif
        }else{

#ifdef DIRECT_POSITION_MODE
                //write position mode
                data[0]= POSITION_MODE;
                data[1]= 0x0000;
                WriteObject(0x6060, 0x00, data, arguments);

                //write max velocity for position_mode
                data[0]= MAX_VELOCITY_DIRECT_POS_MODE;
                data[1]= 0x0000;
                WriteObject(0x607f, 0x00, data, arguments);

                //write max acceleration
                data[0]= MAX_ACCELAERATION_DIRECT_POS_MODE;
                data[1]= 0;
                WriteObject(0x60C5, 0x00, data, arguments);
#else
                //write profiled position mode
                data[0]= PROFILE_POSITION_MODE;
                data[1]= 0x0000;
                WriteObject(0x6060, 0x00, data, arguments);

                //write profile acceleration
                data[0]= ACCELERATION_POSITION_MODE_CAM;
                data[1]= 0x0000;
                WriteObject(0x6083, 0x00, data, arguments);

                //write profile deceleration
                data[0]= DECELERATION_POSITION_MODE_CAM;
                data[1]= 0x0000;
                WriteObject(0x6084, 0x00, data, arguments);

                //write profile velocity
                data[0]= VELOCITY_POSITION_MODE_CAM;
                data[1]= 0x0000;
                WriteObject(0x6081, 0x00, data, arguments);
#endif
        }
        return(0);
}

int go_to_position( int a_fifth_degrees, struct motor_args *arguments, struct msg_queue *queue){
/*!\brief Turns the motor to the desirable position relative to the home position.
 * \param int a_fifth_degrees  : go to position a_fifth_degrees*1/5° relative to the home position
 * \param struct motor_args *arguments  : specific motor parameters
 * \param struct msg_queue *queue  : address from the actual message queue
 * \retval -1 : quick stop command appeared
 * \retval  0 : everything worked fine
 */
    WORD data[2]= {0x0000, 0x0000};

#ifdef DIRECT_POSITION_MODE
    WORD answer;
    WORD *answer_buf;
#else
    WORD ControlwordCommand=0;
    WORD statusword;
#endif

    //calculate target position
    if( arguments->my_queue_addr != ADDRESS_CAMERA_MOTOR ){
        data[0]= a_fifth_degrees*2048/1800;
    }else{
            data[0]= a_fifth_degrees*2048*8/1800;
        }
    data[1]=0;
#ifdef DEBUG_INFO_ADVANCED
    printf("received a_tenth_degrees: %d   go to position:  %d\n",a_fifth_degrees, data[0]);
#endif

#ifdef DIRECT_POSITION_MODE
        WriteObject(0x2062, 0x00, data, arguments);
        //wait until position reached
        do{
            if ( check_if_stop_MQ_for_timeout(10, arguments, queue, 0, 0)==-1   ){ // if quick_stop command appeared return -1
                return(-1);
            }
            pthread_mutex_lock(&mutex_uart);
            ReadObject(0x6064, 0x00, &answer_buf, arguments);
            answer=answer_buf[3];
            pthread_mutex_unlock(&mutex_uart);
        }while( answer != data[0] );
#else
        //profiled position mode
        WriteObject(0x607a, 0x00, data, arguments);
        //enable and start
        ControlwordCommand=ControlwordCommand & 0xfe7F;
        ControlwordCommand=ControlwordCommand | 0x001f;
        data[0]= ControlwordCommand;
        data[1]= 0x00;
        WriteObject(0x6040, 0x00, data, arguments);
        //wait until position reached
        do{
            if ( check_if_stop_MQ_for_timeout(10, arguments, queue, 0, 0)==-1   ){ // if quick_stop command appeared return -1
                return(-1);
            }
            statusword=get_statusword(arguments);
        }while(  (statusword & 0x0400) != 0x0400  );
#endif
    return(0);
}


void go_to_defined_state( struct motor_args *arguments){
    /*! \brief Brings the EPOS control unit to the well-defined state "ready to switch on".
     *   \param struct motor_args *arguments  : specific motor parameters
     */
    WORD data[2]= {0x0000, 0x0000};
    WORD ControlwordCommand=0;

    //write control word --> disable voltage        //state transition 7,9,10,12
    ControlwordCommand=ControlwordCommand & 0x7d;
    data[0]= ControlwordCommand;
    data[1]= 0x00;
    WriteObject(0x6040, 0x00, data, arguments);

    //write control word --> shut down                //state transition 2,6,8
    ControlwordCommand=ControlwordCommand & 0xff7e;
    ControlwordCommand=ControlwordCommand | 0x0006;
    data[0]= ControlwordCommand;
    data[1]= 0x0000;
    WriteObject(0x6040, 0x00, data, arguments);

}


void fault_reset( struct motor_args *arguments){
    /*! \brief Resets a fault in the EPOS control unit, if any.
     * \param struct motor_args *arguments  : specific motor parameters
     */
    WORD *read_buf;
    WORD data[2]= {0x0000, 0x0000};
    WORD answer;

    //read Fault reaction option code
    pthread_mutex_lock(&mutex_uart);
    ReadObject(0x1003, 0x01, &read_buf, arguments);
    answer=read_buf[3];
    pthread_mutex_unlock(&mutex_uart);

        //if error reset it
        if (answer != 0){
#ifdef SHOW_ERROR_INFO
    printf("Received Error code: 0x%x   at addr:   %d\n", answer, arguments->my_queue_addr);
    printf("write control word--> fault reset\n");
#endif
            data[0]= 0x0080;
            data[1]= 0x0000;
            WriteObject(0x6040, 0x00, data, arguments);
        }
}


int check_MQ_and_wait(WORD timeout_in_s, struct motor_args *arguments, struct msg_queue *queue, time_t start_time, WORD centrifugation_time){
/*! \brief Checks the message queue for a quick stop- or gettime command for timeout_in_s seconds.
 * \param WORD timeout_in_s  : time in seconds to wait in function
 * \param struct motor_args *arguments  : specific motor parameters
 * \param struct msg_queue *queue  : message queue ID and local address
 * \param time_t start_time  : start time of the centrifugation process
 * \param WORD centrifugation_time  : expected centrifugation time
 * \retval -1 : quick stop command appeared
 * \retval  0 : everything worked fine
 */
    struct msg_buf MQ_message_struct;

    unsigned char stop[4] = {'s','t','o','p'};
    unsigned char ask_for_time[4] = {'g','e','t','t'};

    WORD data[2]= {0x0000, 0x0000};
    WORD ControlwordCommand=0;

    int received;
    time_t time_diff, time_left, start_time_on_velocity;
    start_time_on_velocity=time(NULL);
    time_diff=0;
    while(  time_diff<timeout_in_s  ){

        received=recv_message_timeout(queue, &MQ_message_struct, 50000);

        if(received != -1) {

            if(memcmp(stop, MQ_message_struct.cmd_id, 4) == 0 ){
#ifdef DEBUG_INFO_BASIC
    printf("do quick stop   addr.: %d \n", arguments->my_queue_addr);
#endif
                ControlwordCommand=ControlwordCommand & 0x7B;
                ControlwordCommand=ControlwordCommand | 0x02;
                data[0]= ControlwordCommand;
                data[1]= 0x00;
                WriteObject(0x6040, 0x00, data, arguments);

                MQ_message_struct.param[0] = STOPPED;
                MQ_message_struct.cmd_param_len=1;
#ifdef DEBUG_INFO_BASIC
    printf(" quick stop done  addr.: %d \n", arguments->my_queue_addr);
#endif
                send_message(queue, &MQ_message_struct, ADDR_COMD);
                return (-1);
            }else{
            if(memcmp(ask_for_time, MQ_message_struct.cmd_id, 4) == 0){
#ifdef DEBUG_INFO_BASIC
    printf("asked for time addr.: %d \n", arguments->my_queue_addr);
#endif

                time_left=centrifugation_time-difftime(time(NULL), start_time);
                if(time_left<0){time_left=0; }

                MQ_message_struct.cmd_param_len=3;
                MQ_message_struct.param[0] = 'T';
                MQ_message_struct.param[1] = time_left & 0x00ff;
                MQ_message_struct.param[2] = (time_left & 0xff00)>>8;
                send_message(queue, &MQ_message_struct, ADDR_COMD);
                }else{
                        //if other message send BUSY
                        MQ_message_struct.param[0] = BUSY;
                        MQ_message_struct.cmd_param_len=1;
                        send_message(queue, &MQ_message_struct, ADDR_COMD);
                    }
            }
        }
        usleep(31000);
        time_diff=difftime(time(NULL), start_time_on_velocity);
    }
    return(0);
}

