/*! \file main.c
\brief \M Main function.<br>
 * \author Christian Affolter, nanoTRONIC
 * \date Nov 2009
*/


/*! \mainpage motord(motor demon)
 *
 * A program to control three maxon motors.
 *
 * The motors are driven over the maxon control unit EPOS2 Module 36/2. EPOS2 Module 36/2 is a positioning controller from maxon motor,
 * which provide miscellaneous functions to control a maxon motor.
 *
 * The motord communicates over the RS232 interface with the first EPOS control unit, which controls the camera motor. The next two EPOS
 * units are connected over a CAN-bus with the first control unit. The first EPOS control unit sends all the commands which doesn't have the own
 * Node-ID, from the RS232 interface over the CAN-bus to the other EPOS control units.
 *
 * The motord creates for each motor a thread so it can control the motors in "parallel" mode.
 *
 * The motord is guided from the command demon(comd). The comd communicates with the program over a message queues. For each motor thread a message
 * queue is created. The motorthreads are waiting on commands in the message queue. If a command is received the thread is executing the corresponding
 * command or send a BUSY back to the comd. If a command was executed the motord sends the corresponding answer back to the comd.
 * \author Christian Affolter, nanoTRONIC
 * \date Nov 2009
 *
*/


/*! \page page1 The motord consists of the following files
 *
 * \section sec11 Files and their function
 *
 *  <b> - main.c </b>
 *
 *       - main() : initializes and creates the motorthreads and opens the connection to the EPOS control units.
 *
 *  <b> - motor_thread.c </b>
 *
 *        - motor_thread() : controls a Maxon motor.
 *
 * <b> - init_parameter.c </b>
 *
 *        - init_parameter() : initializes and stores the parameters of the motors.
 *
 * <b> - functions.c </b>
 *          Advanced functions to control a Maxon motor with an EPOS control unit.
 *          It contains the following functions:
 *
 *        - check_if_stop_MQ_for_timeout() : checks the message queue for a quick stop- or getTime command.
 *        - get_statusword() : reads the statusword from an EPOS control unit.
 *        - go_to_velocity() : drives the motors to a desirable velocity.
 *        - go_to_home_position() : drives the motors to the home position.
 *        - go_to_position() : turns the motor to a desirable position relative to the home position.
 *        - go_to_defined_state() : brings the EPOS control unit to the well-defined state "ready to switch on"
 *        - fault_reset() : resets a fault in the EPOS control unit, if any.
 *        - check_MQ_and_wait() : checks the message queue for a quick stop- or gettime command for timeout_in_s seconds.
 *
 * <b> - epos.c </b>
 *          A GNU/Linux C library to control an EPOS motor control unit by maxon motor. It contains the following functions:
 *
 *        - endian_swap16() : swaps a WORD from little- to high endian or vice versa.
 *        - openEPOS() : opens the connection to EPOS.
 *        - checkEPOSerror() : checks the global variable E_error for EPOS error code.
 *        - writeBYTE() : writes a single BYTE to EPOS.
 *        - writeWORD() : writes a single WORD to EPOS.
 *        - readBYTE() : reads a single BYTE from EPOS, timeout implemented.
 *        - readWORD() : reads a single WORD from EPOS, timeout implemented.
 *        - CalcFieldCRC() : calculates the CRC (cyclic redundancy check) copied from EPOS Communication Guide, p.8.
 *        - sendCom() : sends a command to EPOS, taking care of all necessary 'ACK' and checksum test.
 *        - readAnswer() : reads an answer frame from EPOS.
 *        - ReadObject() : reads an entry from an EPOS register.
 *        - WriteObject() : writes an entry to an EPOS register.
 *
 *
 * <b> - msg_queue.c </b>
 *
 *          Functions to use the message queue. It contains the following functions:
 *
 *        - print_queue_status() : prints the queue status.
 *        - establish_queue() : establishes the message queue.
 *        - remove_queue() : removes the message queue.
 *        - connect_to_queue() : connects to the message queue.
 *        - send_message() : puts a message into the message queue.
 *        - recv_message_timeout() : is waiting for a message in the message queue for a timeout.
 *        - recv_message() : is waiting for a message in the message queue.
 *
 *
 * <b> - global_def.h </b>
 *
 *          Global definitions for the program.
*/

/*! \page page2 The usage of the message queue
 *  \section sec21 Establish the queue
 *
 *  	The comd generates the message queue for the communication between the comd and the motord with the function call establish_queue().
 *  	Each motor thread can then connect to the queue with the function call connect_to_queue(). In this call the struct msg_queue is written, where the
 *  	unique message queue ID and the local message queue address is stored.
 *  	After this initialization it is possible to use the functions recv_message() and send_message() for the communication inbetween these program parts.
 *  	The format of the data which will get exchanged is defined in the struct msg_buf.
 *
 *
 * \section sec22 Usage of the message queue
 *  	First each motorthread is waiting in the function motor_thread() for a command in the message queue. When it receives a command it will then execute it.
 *  	During the execution of a command it is possible to send a quick stop command, which stops the corresponding
 *  	motor immediately. During a centrifugation process it is possible to ask for the remaining centrifugation time. If an other command is received the motord
 *  	will send a "BUSY" back. When a command is sucessfully executed the motord will send an "ACK" back. If a quick stop command occured the motord sends a "STOPPED" back.
 *
 *
 * \section sec23 The commands that can get executed
 *
 *  	\subsection sub1 Initialization
 *
 *			\subsubsection subsub11 Command identifier stored in the struct msg_buf
 *           <table border>
 *              <tr>
 *                 <td><b> element in msg_buf : </b></td>
 *                 <td><b> Value : </b></td>
 *              </tr>
 *              <tr>
 *                 <td> cmd_id[4] </td>
 *                 <td> "init" </td>
 *			    </tr>
 *			 </table>
 *
 *			\subsubsection subsub12 Description of the command
 *			  The initialization is calling three functions.
 *
 *			  - The first function call fault_reset(), resets a possible fault in the EPOS control unit.
 *			  - Then the function call init_parameter(), where all the parameters for the EPOS control unit are getting stored.
 *			  - At last the function call go_to_home_position(), makes sure that the motors have a defined position.
 *
 *  	\subsection sub2 Homing
 *
 *			\subsubsection subsub21 Command identifier stored in the struct msg_buf
 *           <table border>
 *              <tr>
 *                 <td><b> element in msg_buf: </b></td>
 *                 <td><b> Value: </b></td>
 *              </tr>
 *              <tr>
 *                 <td> cmd_id[4] </td>
 *                 <td> "home" </td>
 *			    </tr>
 *			 </table>
 *
 *			\subsubsection subsub22 Description of the command
 *			 The homing function uses a provided function from the EPOS controller unit, which detects the light barrier and
 *			 moves then to the next index from the motor. The index from the motor is a defined position within one rotation.
 *
 *  	\subsection sub3 Centrifugation
 *			\subsubsection subsub31 Command identifier stored in the struct msg_buf
 *
 *           <table border>
 *              <tr>
 *                 <td><b> element in msg_buf: </b></td>
 *                 <td><b> Value: </b></td>
 *              </tr>
 *              <tr>
 *                 <td> cmd_id[4] </td>
 *                 <td> "cent" </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[0] </td>
 *			       <td> acc1 </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[1] </td>
 *			       <td> acc2 </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[2] </td>
 *			       <td> velocity1 high byte </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[3] </td>
 *			       <td> velocity1 low byte </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[4] </td>
 *			       <td> velocity2 high byte </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[5] </td>
 *			       <td> velocity2 low byte </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[6] </td>
 *			       <td> time on velocity1 high byte </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[7] </td>
 *			       <td> time on velocity1 low byte </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[8] </td>
 *			       <td> time on velocity2 high byte </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[9] </td>
 *			       <td> time on velocity2 low byte </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[10] </td>
 *			       <td> deceleration </td>
 *			    </tr>
 *          </table>


 *		\subsubsection subsub32 Description of the command
 *		  The centrifugation is executing the following centrifugation profile:
 *
 *  		
 *
		 *   <table border>
		 *  <tr>
		 *     <td><b> Specifier: </b></td>
		 *     <td><b> Unit: </b></td>
		 *     <td><b> Definition: </b></td>
		 *  </tr>
		 *  <tr>
		 *     <td> acc1 </td>
		 *     <td> [rad/s²] </td>
		 *     <td> Acceleration to reach velocity 1 </td>
		 *  </tr>
		 *  <tr>
		 *     <td> acc2 </td>
		 *     <td> [rad/s²] </td>
		 *     <td> Acceleration to reach velocity 2 </td>
		 *  </tr>
		  *  <tr>
		 *     <td> dec </td>
		 *     <td> [rad/s²] </td>
		 *     <td> Deceleration for the centrifugation profile </td>
		 *  </tr>
		 *  <tr>
		 *     <td> velocity 1 </td>
		 *     <td> [rev/min] </td>
		 *     <td> 1st velocity step </td>
		 *  </tr>
		 *  <tr>
		 *     <td> velocity 2 </td>
		 *     <td> [rev/min] </td>
		 *     <td> 2nd velocity step </td>
		 *  </tr>
		 *  <tr>
		 *     <td> time 1 </td>
		 *     <td> [s] </td>
		 *     <td> Time on velocity 1 </td>
		 *  </tr>
		 *  <tr>
		 *     <td> time 2 </td>
		 *     <td> [s] </td>
		 *     <td> Time on velocity 2 </td>
		 *  </tr>
		 *  </table>
 *
*		  The centrifugation is calling three functions.
*
*		 - go_to_velocity() and check_MQ_and_wait() for the execution of the centrifugation profile.
*		 - At the end of the centrifugation profile go_to_home_position(), to make sure that the motor has a defined position.
 *
 *  	\subsection sub4 Move to a position relative to the home position
 *			\subsubsection subsub41 Command identifier stored in the struct msg_buf
 *           <table border>
 *              <tr>
 *                 <td><b> element in msg_buf: </b></td>
 *                 <td><b> Value: </b></td>
 *              </tr>
 *              <tr>
 *                 <td> cmd_id[4] </td>
 *                 <td> "move" </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[0] </td>
 *			       <td> position high byte </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[1] </td>
 *			       <td> position low byte </td>
 *			    </tr>
 *			 </table>
 *
 *			\subsubsection subsub42 Description of the command
 *			The move command turns the motor to the desirable position relative to the home position. The resolution is one-fifth degree.
 *
 *  	\subsection sub5 Quick stop
 *			\subsubsection subsub51 Command identifier stored in the struct msg_buf
 *           <table border>
 *              <tr>
 *                 <td><b> element in msg_buf: </b></td>
 *                 <td><b> Value: </b></td>
 *              </tr>
 *              <tr>
 *                 <td> cmd_id[4] </td>
 *                 <td> "stop" </td>
 *			    </tr>
 *			 </table>
 *			\subsubsection subsub52 Description of the command
 *
 *			The stop command is executing the provided function from the EPOS controller quick stop. It stops the motors immediately.
 *
 *  	\subsection sub6 Ask for the remaining centrifugation time
 *			\subsubsection subsub61 Command identifier stored in the struct msg_buf
 *           <table border>
 *              <tr>
 *                 <td><b> element in msg_buf: </b></td>
 *                 <td><b> Value: </b></td>
 *              </tr>
 *              <tr>
 *                 <td> cmd_id[4] </td>
 *                 <td> "gett" </td>
 *			    </tr>
 *			 </table>
 *
 *			\subsubsection subsub62 Description of the command
 *
 *			The gett command calculates the remaining centrifugation time.
 *			The calculated time will get sent back through the message queue like it is shown in the following table:

 *           <table border>
 *              <tr>
 *                 <td><b> element in msg_buf: </b></td>
 *                 <td><b> Value: </b></td>
 *              </tr>
 *              <tr>
 *                 <td> cmd_id[4] </td>
 *                 <td> "gett" </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[0] </td>
 *			       <td> 'T' </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[1] </td>
 *			       <td> remaining centrifugation time low byte </td>
 *			    </tr>
 *			    <tr>
 *			       <td> param[2] </td>
 *			       <td> remaining centrifugation time hight byte </td>
 *			    </tr>
 *			 </table>
 *
 *
 *
*/










#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>  //data types like pthread_t
#include <pthread.h>

#include "global_def.h"
#include "motor_thread.h"
#include "epos.h"

#ifdef MICRO_C_DEF
    const char ttyUL[] = "/dev/ttyS1";
#else
    const char ttyUL[] = "/dev/ttyUSB0";
#endif


int main(void) {
    /*!
    \brief main() initializes and creates the motor threads and opens the connection to the EPOS control units.
    \author Christian Affolter, nanoTRONIC
    \date Oct 2009
    */
    pthread_t tid_mot1, tid_mot2, tid_mot3;
    pthread_attr_t thread_attr;

#ifdef DEBUG_INFO_BASIC
    printf("-- motord, built: %s, %s --\n", __TIME__, __DATE__);
#endif

    struct motor_args arg3 = {3, "CAM motor 3", ADDRESS_CAMERA_MOTOR};
    struct motor_args arg1 = {1, "motor 1", ADDRESS_1_CENTRIFUGE_MOTOR};
    struct motor_args arg2 = {2, "motor 2", ADDRESS_2_CENTRIFUGE_MOTOR};

    pthread_attr_init(&thread_attr);
    pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);

    if( openEPOS(ttyUL) < 0 ) {
        fprintf(stderr, "ERROR: Can't initializing tty device %s!\n", ttyUL);
        return(-1);
    }

    pthread_create(&tid_mot3, &thread_attr, (void*) motor_thread, (void *) &arg3);
    sleep(20);
    pthread_create(&tid_mot1, &thread_attr, (void*) motor_thread, (void *) &arg1);
    sleep(20);
    pthread_create(&tid_mot2, &thread_attr, (void*) motor_thread, (void *) &arg2);


    while(1) {
        sleep(10);
    }

}

